#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import rpy2.robjects as robjects
from rpy2.robjects.packages import importr

"""
Classes to interface in a limited way (presets) with 
R's packages 'topGO' and 'GOstats' via rpy2.
"""

#TODO: Re-write these classes without using the 'robject.globalenv'
#      because that leads to issues, when "crossing" objects from 
#      different methods. Add some input validation.

# ==============================================================================

class KEGGer(object):


    def __init__(self, keggmap, alpha, organism):

        self.mappings = keggmap
        self.alpha = alpha
        self.organism = organism

        KEGGdb = importr('KEGG.db')
        GOstats = importr('GOstats')
        GSEABase = importr('GSEABase')

    def perform_kegg_enrichment(self, targetspath):

        """
        Uses the R GOstats package to calculate KEGG pathway enrichment for a
        suplied list of genes/transcripts of interest, saving results to file.
        """

        robjects.globalenv["mappings"] = self.mappings
        robjects.globalenv["alpha"] = self.alpha
        robjects.globalenv["organism"] = self.organism


        read_target_group_of_interest(targetspath)

        handle_result_saving(targetspath, "keggenrich/")

        robjects.r('''tryCatch(
                    {
                      KEGGmap <- read.csv(mappings, sep="\t", colClasses=
                                    c("character", "character"), header=FALSE)
                      colnames(KEGGmap) <- c("path_id", "gene_id")
                      keggFrame=KEGGFrame(KEGGmap,organism=organism)
                      gsc <- GeneSetCollection(keggFrame, setType = 
                                    KEGGCollection())
                      universe <- unique(KEGGmap[,2])
                      universe <- unlist(as.character(universe))

                      kparams <- GSEAKEGGHyperGParams(
                                  name="Custom GSEA based params",
                                  geneSetCollection=gsc,
                                  geneIds = targetset,
                                  universeGeneIds = universe,
                                  pvalueCutoff = alpha,
                                  testDirection = "over")

                      kOver <- hyperGTest(kparams)
                      enrichRes <- summary(kOver)

                      write.table(enrichRes, file = outname, sep = "\t", 
                         row.names = FALSE, quote = FALSE)

                    }, error = function(e){ print( paste(basename, 
                    " failed to have any KEGG pathways mapped to its members!") 
                    ) }
                    )
        ''')


# ==============================================================================

class GOrich(object):

    # NOTE: alpha parameter is not currently been used, here.
    #       Check possibily of using it as a filter later.
    def __init__(self, gomap, ontology='BP', alpha=0.01):

        self.mappings = gomap
        self.ontology = ontology
        self.alpha = alpha

        topGO = importr('topGO')
        JSONlite = importr('jsonlite')

    def perform_go_enrichment(self, targetspath):

        """
        Uses the R topGO package to calculate GO term enrichment in suplied
        lists of genes/transcripts of interest, saving results to file.
        """

        robjects.globalenv["mappings"] = self.mappings
        robjects.globalenv["ontology"] = self.ontology
        robjects.globalenv["alpha"] = self.alpha

        read_target_group_of_interest(targetspath)
        handle_result_saving(targetspath, "goenrich/"+self.ontology)

        robjects.r('''
        capture.output(id2go <- readMappings(file = mappings), file="/dev/null")
        transcriptNames <- names(id2go)

        interestingGenes <- factor(as.integer( 
                        transcriptNames %in% as.character(unlist(targetset)) ) )
        names(interestingGenes) <- transcriptNames

        success = FALSE
        ''')

        robjects.r('''
                      tryCatch(
                      {
                      capture.output(
                        GOdata <- new("topGOdata", ontology = ontology,
                        allGenes = interestingGenes, annot = annFUN.gene2GO, 
                        gene2GO = id2go)
                      , file="/dev/null" )
                      success <- TRUE
                      }, 
                      error = function(e){ 
                                print( paste(basename, 
                                "failed to have any GO", ontology, 
                                "terms mapped to its members!") )
                                success <- FALSE 
                                }
                      )
                   ''')

        robjects.r('''
                if (success == TRUE){
                    capture.output(
                        resultFisher <- runTest(GOdata, algorithm = "classic",
                        statistic = "fisher")
                    , file="/dev/null" )

                    capture.output(
                        resultFisher.Elim <- runTest(GOdata, algorithm = "elim",
                        statistic = "fisher")
                    , file="/dev/null" )

                    enrichRes <- GenTable(GOdata, classicFisher = resultFisher,
                               elimFisher = resultFisher.Elim, 
                               orderBy = "elimFisher", ranksOf ="classicFisher",
                               topNodes = 30)

                    write.table(enrichRes, file = outname, sep = "\t", 
                                row.names = FALSE, quote = FALSE)
                }
                   ''')


    def extract_full_annot_mapping(self, genelist, pathout):

        """
        Extracts list of GO term annotations (ancestors included)
        into a psedo-json (.js w/ 'annotmap' var file.
        """

        robjects.globalenv["mappings"] = self.mappings

        robjects.globalenv["targetset"] = robjects.StrVector(genelist)
        robjects.globalenv["pathout"] = pathout


        robjects.r('''
        capture.output(id2go <- readMappings(file = mappings), file="/dev/null")
        transcriptNames <- names(id2go)

        interestingGenes <- factor(as.integer( 
                        transcriptNames %in% as.character(unlist(targetset)) ) )
        names(interestingGenes) <- transcriptNames

        success = FALSE
        onts <- c("BP", "MF", "CC")
        comboList <- list()
        ''')

        robjects.r('''
                  tryCatch(
                        {
                        for (ont in onts){
                            capture.output(
                                GOdata <- new("topGOdata", ontology = ont,
                                              allGenes = interestingGenes, 
                                              annot = annFUN.gene2GO, 
                                              gene2GO = id2go)
                            , file="/dev/null" )

                            go2ids <- genesInTerm(GOdata, usedGO(GOdata))
                            for (go in names(go2ids)){
                                comboList[[go]] <- go2ids[[go]]
                            }
                        }
                        success <- TRUE
                        }, 

                        error = function(e){
                                  print(e) 
                                  print( paste(basename, 
                                  "failed to have any GO", ont, 
                                  "terms mapped to its members!") )
                                  success <- FALSE 
                                  }
                  )
              ''')

        robjects.r('''
                if (success == TRUE){
                    jsonout = toJSON(comboList, pretty = TRUE)
                    fh <- file(pathout)
                    writeLines(paste("var annotmap = ", jsonout), fh)
                    close(fh)
                }
                   ''')


# accessory py2r helpers
# ------------------------------------------------------------------------------

def read_target_group_of_interest(targetspath):

    """
    Reads a tsv file, extracts the first column (while assuming they
    are transcript/gene identifiers) and sets them as a vector in R
    global environament.
    """

    genes = []
    with open(targetspath) as fh:
        for line in fh:
            token = line.split("\t")
            genes.append(token[0].strip())

    robjects.globalenv["targetset"] = robjects.StrVector(genes)


def handle_result_saving(targetspath, enrich):

    """
    Creates (if necessary) output folder to save enrichment results
    and 'returns' full path to the results file being saved.
    """

    # SAVE IT TO DISK
    basename = os.path.splitext(os.path.basename(targetspath))[0]
    foldername = os.path.dirname(targetspath)

    robjects.globalenv["basename"] = basename

    outname = os.path.join(foldername, enrich, basename+"_enrichment.tsv")
    robjects.globalenv["outname"] = outname

    # if not exists, create folder to save enrichment results
    savepath = os.path.join(foldername, enrich)
    try: 
        os.makedirs(savepath)
    except OSError:
        if not os.path.isdir(savepath):
            raise


#!/usr/bin/python
# -*- coding: utf-8 -*-

import os, re, json
import pandas as pd

pd.set_option('display.max_colwidth', -1)
pd.options.mode.chained_assignment = None  # default='warn'


class Slicer(object):

    def __init__(self, mapfile):

        self.descmap = self.read_annotation_file(mapfile, False)


    # useful for simple 2 column 1-to-1 (key:value) annotation files
    def read_annotation_file(self, f, list_=True):

        '''
        Reads and parses a tsv annotation file into 'TWO columns' split:
        identifier(key):annotations(values). -- all available annotations.
        Note: Expects _whitespace_, _comma_ or _semi-colon_ as separator.
        '''

        # TODO: increment/add input sanitization

        annotDict = {}
        patt = re.compile('([^,;\s]+)[,;\t\v]+(.*)', re.IGNORECASE)

        try:
            with open(f, 'r') as fh:
                for line in fh:
                    m = patt.search(line)
                    if m:
                        annotDict.setdefault(m.group(1), 
                                                []).append(m.group(2).strip())
                    else:
                        # considers that row has an identifier only
                        annotDict.setdefault(line.strip(), []).append('')
        except IOError:
            print 'Could not read file:', f
            raise

        if list_ == False:
            for key in annotDict:
                annotDict[key] = ' '.join(annotDict[key])

        return annotDict


    def writeout_additional_annotation(self, inpath, outpath):

        '''
        Creates new copy of file with added annotations provided
        via mapping dictionary.
        '''

        try:
            with open(inpath, 'r') as fh:
                lines = fh.readlines()

            with open(outpath, 'w') as out:
                for line in lines:
                    frags = line.split('\t')
                    try:
                        #print self.descmap[frags[0]]
                        out.write(line.strip('\n')+'\t'+
                                    self.descmap[frags[0].strip()]+'\n')
                    except KeyError:
                        out.write(line.strip('\n')+'\t\n')
                        raise
        except IOError as err:
            print 'Could not read/write file:', err.filename
            raise


    # NOTE: maybe make it more 'path flexible'
    def process_enrichment_values(self, directory, basename, alpha):

        '''
        Processes KEGG/GO enrichment results files (produced by 
        GOstats and topGO) into a single dictionary.
        '''

        try:
            bp = self.read_enrichment_tsv(os.path.join(directory, 'goenrich/BP', 
                                  basename+'_enrichment.tsv'), 
                                  'GO Biological Process')
        except:
            bp = None
            
            
        try:
            mf = self.read_enrichment_tsv(os.path.join(directory, 'goenrich/MF', 
                        basename+'_enrichment.tsv'), 'GO Molecular Function')
        except:
            mf = None
            
        try:
            cc = self.read_enrichment_tsv(os.path.join(directory, 'goenrich/CC',
                        basename+'_enrichment.tsv'), 'GO Cellular Component')
        except:
            cc = None
            
            
        try:
            kegg = self.read_enrichment_tsv(os.path.join(directory,'keggenrich', 
                                basename+'_enrichment.tsv'), 'KEGG Pathways')
        except:
            kegg = None


        extra = {'bp':bp, 'mf':mf, 'cc': cc, 'kegg':kegg, 'alpha': alpha}

        return extra


    def read_enrichment_tsv(self, tsvpath, ont):

        '''
        Reads enrichment results tsv files (as produced either by GOstats or
        the topGO R packages) into pandas dataframes (including the headers).
        '''

        try:
            if ont != 'KEGG Pathways':
                df = pd.read_csv(tsvpath, sep='\t',
                    converters={'elimFisher': lambda x: parse_ev(x)})
                df[['elimFisher']] = df[['elimFisher']].apply(pd.to_numeric)
            else:
                df = pd.read_csv(tsvpath, sep='\t', 
                                    converters={'KEGGID': lambda x: str(x)})
        except:
            f = os.path.splitext(os.path.basename(tsvpath))[0]
            print 'No %s enrichment found for %s!' % (ont, f)
            raise

        return df


    def generate_support_js_file(self, annots, outdir, f, varname):

        '''
        Turns a given dictionary and creates mock-json files
        (js files with a single var) to be used by the html 
        enrichments reports.
        '''
        outfile = os.path.join(outdir, f)

        if os.path.isfile(outfile):
            return
        else:
            try:
                with open(outfile, 'w') as fh:
                    fh.write('var '+varname+' = ')
                    jsondump = json.dumps(annots)
                    fh.write(jsondump)
            except IOError:
                print 'Could not read file:', f
                raise

        return outfile







def parse_ev(val):

    '''
    Identifies and converts '< 1e-30' strings
    into numeric convertable strings
    '''

    patt = re.compile('^<\s+(.+)')
    m = patt.search(val)
    if m:
        return m.group(1)
    else:
        return val


#!/usr/bin/python
# -*- coding: utf-8 -*-

import networkx as nx
import pandas as pd
import os, math, subprocess, tempfile, misc

from libraries.ontology import OBOOntology

pd.set_option('display.max_colwidth', -1)
pd.options.mode.chained_assignment = None  # default='warn'

class OBOe(object):

    """
    This class is used to create SVG GO sub-graphs
    with overlaid color-coded enrichment statistics
    and/or PNG/MAP files.

    INPUT:

    - Ontology OBO file
    - topGO result files

    OUTPUT:

    - dot, svg (and optionally other graphic filetypes)

    """


    def __init__(self, obopath):

        self.obofile = obopath
        self.ontology = self.read_obo()


    def read_obo(self):

        ont = OBOOntology(self.obofile)
        return ont

    def create_ntx_graph(self, termlist):

        gx = self.ontology.to_networkx(termlist)
        return gx

    def ntx_nodes(self, ntxgraph):

        return ntxgraph.nodes();


    def get_directed_edges(self, termlist):

        """
        Takes full list of terms comprising a GO
        sub-graph and return its directed 'is_a'
        edges and the root term (the only sink).
        """

        root = ""
        edgelist = []

        clonelist = [x for x in termlist]
        terms = []

        for term in clonelist:
            par_edges = self.ontology.parent_edges(term)
            # if term has any parent...
            if len(par_edges) > 0:
                for pe in par_edges:
                    if pe[0] == "is_a":
                        edgelist.append( ( term, pe[1].id) )
            else:
                root = term

        return edgelist, root


    def create_basic_dag(self, edgelist):

        dag = nx.DiGraph()
        dag.add_edges_from(edgelist)

        return dag


    def generate_basic_dot(self, dag):

        """
        Converts networkx graph into a dot
        file contents output (as a list).
        """

        P = nx.nx_pydot.to_pydot(dag)
        A = nx.nx_agraph.to_agraph(dag)

        temp = tempfile.NamedTemporaryFile()
        A.write(temp.name)

        dot = []
        for line in temp:
            dot.append(line.strip())

        return dot


    def read_enrichment_tsv(self, filepath):

        df = pd.read_csv(filepath, sep="\t",
                        converters={'elimFisher': 
                            lambda x: misc.parse_ev(x)})
        df[['elimFisher']] = df[['elimFisher']].apply(pd.to_numeric)
        return df


    def extract_go_terms(self, df, alpha):

        """
        Tries to extract a list of GO accessions ids at a given
        statistical significance from a topGO results dataframe.
        """

        try:
            fatia = df["GO.ID"].loc[df['elimFisher'] < float(alpha)]
            subset = [x for x in fatia]
            return subset
        except:
            return False

    # -------------------------------------------------------------------------


    def enhance_dot(self, outname, nodes, listdotedges, df, alpha, root):

        #temp = tempfile.NamedTemporaryFile(delete=False)

        outfile = outname+".dot"
        temp = open(outfile, "w")

        sig = self.extract_go_terms(df, alpha)

        # dot file header
        temp.write('strict digraph  {\n graph[rankdir="BT"];\n')

        added = [] # stores list of nodes that are customized in the dot

        for index, row in df.iterrows():
            if row["elimFisher"] < alpha:
                # set root term as sink
                if row["GO.ID"] == root:
                    # get term name here    
                    temp.write('"'+row["GO.ID"]+'" '+'[rank="sink"];\n')
                    added.append(row["GO.ID"])
                else:

                    term_name = self.ontology[row["GO.ID"]].name
                    truncterm = self.process_term_name(term_name)
                    trunctermplus = row["GO.ID"]+"\n"+truncterm+"\n"+ \
                                    str(row["elimFisher"])+"\n"+"("+ \
                                    str(row["Significant"])+"/"+ \
                                    str(row["Annotated"])+")"

                    color_hsv = self.get_enrichment_color(row["elimFisher"])

                    tooltip = term_name
                    temp.write('"'+row["GO.ID"]+'" [ id="'+row["GO.ID"]+
                               '" shape="note" style="filled" color="'+
                               str(color_hsv[0])+' '+str(color_hsv[1])+' '+
                               str(color_hsv[2])+'" label="'+trunctermplus+
                                                '" tooltip="'+tooltip+'"];\n')

                    added.append(row["GO.ID"])


        # Handle statistically non-significant nodes
        for go in nodes:
            if go not in added:
                if go == root:
                    # get term name here
                    term_name = self.ontology[go].name
                    temp.write('"'+go+'" '+'[rank="sink" style="filled" label="'
                                                    +go+'\\n'+term_name+'"];\n')
                else:
                    term_name = self.ontology[go].name
                    temp.write('"'+go+'" '+'[style="filled" label="'+go+
                                           '" tooltip="'+term_name+'"];\n')

        for edge in listdotedges[1:]:
            temp.write(edge+"\n")
        temp.close()

        return temp.name


    def export_dag_as(self, dotfile, filetype):

        """
        Exports a dag (dot) file as either an SVG or PNG/MAP
        """

        outdef = ()

        if filetype == "svg":
            outdef = (".svg", "-Tsvg")
        elif filetype == "png":
            outdef = (".png", "-Tpng")
        elif filetype == "map":
            outdef = (".map", "-Tcmapx")
        else:
            print "Filetype ", filetype, " not recognised!"
            sys.exit()

        outfile = os.path.splitext(dotfile)[0]+outdef[0]
        with open(outfile,'w') as s_out:
            subprocess.call(['dot', outdef[1], dotfile], stdout=s_out)


    def process_term_name(self, name):

        """
        Processes a 'term name' in order to obtain a 
        truncated version to better fit a graphviz node.
        """

        if len(name) > 20:

            trunc_name = ''
            token = name.split(' ')

            # if 1st word has MORE than 20 characters
            if len(token[0]) > 20:
                # RETURN the first 18 characters plus '...'
                return token[0][0:17] + '...'

            else:
                last = ""
                for e in token:
                    # if ANY subsequent SINGLE word(s) has 
                    # MORE than 20 characters...
                    if len(e) > 20:
                        diff = 20 - (len(last) + len(e))
                        return trunc_name + last + e[0:diff] + '...'

                    else:
                        # if ANY subsequent SINGLE LINE has
                        # MORE than 20 characters...
                        if (len(last) + len(e)) < 20:

                            last = last + " " + e
                            if e == token[-1]:
                                return trunc_name + last

                        else:
                            trunc_name += last + '\\n'
                            last = e
                            if e == token[-1]:
                                return trunc_name + last

                return trunc_name

        # OTHERWISE just return it... 
        else:
            return name


    def get_enrichment_color(self, pvalue):

        """
        Creates an hue for a p-value
        within a chosen HSV color range.
        """

        # hard-coded range
        # TODO: improve this, possibly make it customizable
        #       or make a few sensible presets available
        # pall_1 = (170, 240)
        max_colour = 240
        min_colour = 170

        if pvalue == 0:
            hue = max_colour/float(360)
        else:
            hue = ( min_colour + abs(math.log(pvalue)) ) / float(360)

        return (hue, 1, 1)


from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='makimono',
      version='0.1.3',
      description='Visualizations for RNA-Seq expression data',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GPL v3 License',
        'Programming Language :: Python :: 2.7',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords='GOldwasher GO term enrichment visualization',
      url='http://github.com/hpsbastos/GOldwasher',
      author='hpbastos',
      author_email='hpb29@cam.ac.uk',
      license='GPLv3',
      packages=['GOldwasher'],
      install_requires=[
          'networkx',
          'jinja2',
          'numpy',
          'pandas',
          'matplotlib',
          'rpy2',  
      ],
      scripts=['bin/makisu'],
      test_suite='nose.collector',
      tests_require=['nose', 'nose-cover3'],
      include_package_data=True,
      zip_safe=False)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import pandas as pd
import re, json, pkg_resources

from jinja2 import Template
from decimal import *

pd.set_option('display.max_colwidth', -1)
pd.options.mode.chained_assignment = None  # default='warn'


class Templater(object):

    def __init__(self, title, annot, org, links=None):

        self.title = title
        self.annots = annot

        self.svgpaths = {'bp': 'svg/'+title+'_enrichment_BP.svg', 
                         'mf': 'svg/'+title+'_enrichment_MF.svg', 
                         'cc': 'svg/'+title+'_enrichment_CC.svg' }

        if links == None:
            inserts = json.loads(pkg_resources.resource_string('GOldwasher', 
                                                            '/organisms.json'))

        else:
            with open(links, "r") as f:
                inserts = json.load(f)

        try:
            self.link1 = inserts[org]['insertlink1']
            self.link2 = inserts[org]['insertlink2']
        except:
            print "Are your target organism's linkouts in 'organisms.json' ?"
            # creates fallback non-linking fake anchors
            self.link1 = """'<a href="#null" alt="'+dialogcontent[t]+'">'
                             +dialogcontent[t]+'</a> '+descript+'</br>';"""

            self.link2 = '''<a name="{{ id }}" href="#null" alt="{{ id }}">
                            {{ id }}</a>'''



    # NOTE: currently this method overrides "process_enrichment_dict()" 
    #       regarding the display of missing (None) enrichment info.
    def render_main_page(self, plus = None):

        """
        Renders the main page template for a plot and association expression
        information: gene/transcript list, annotations, (GO/KEGG) enrichment
        results.     
        """

        css_resources = '''
          <link href="support/custom.css" rel="stylesheet" 
                                          type="text/css" />
          <link href="support/tab-content/tabcontent.css" 
                                          rel="stylesheet" type="text/css" />
          <link href="support/jquery/jquery-ui.css" 
                                          rel="stylesheet" type="text/css"/>
        '''


        js_resources = '''
          <!-- These following 2 files need to be generated per project -->
          <script type="text/javascript" src="support/annotations.js">
          </script>
          <script type="text/javascript" src="support/descriptions.js">
          </script>


          <!-- Static support (libraries) files -->
          <script src="support/sorttable.js" type="text/javascript"></script>

          <script src="support/tab-content/tabcontent.js"
          type="text/javascript"></script>

          <script  src="support/svg-pan-zoom/svg-pan-zoom.js" 
          type="text/javascript"></script>

          <script src="support/jquery/jquery-1.10.2.js" type="text/javascript">
          </script>
          <script src="support/jquery/jquery-ui.js" type="text/javascript" >
          </script>

           '''

    # -------------------------------------------------------------------------

        if plus is not None:

            bp = plus['bp']
            mf = plus['mf']
            cc = plus['cc']
            kegg = plus['kegg']

            # a small hack to pass along the significance 
            # level (filter for GO enrichments)
            alpha = plus['alpha']

            # TODO: testing...    
            #jannots = plus['annots']
            jgenes = plus['genes']

        else:
            bp = None
            mf = None
            cc = None
            kegg = None
            alpha = 0.05
            # TODO: test this/alternatives
            jgenes = None


        # TODO: testing...
        if jgenes is not None:

            bigjson = '''
            <script type="text/javascript">

            function toggle(target) {
                var ele = document.getElementById("toggle"+target);
                var text = document.getElementById("display"+target);
                if(ele.style.display == "block") {
                        ele.style.display = "none";
                    text.innerHTML = "show";
                }
                else {
                    ele.style.display = "block";
                    text.innerHTML = "hide";
                }
            } 


            $(function() {
                $(".dialog").dialog({
                    autoOpen: false
                    });
         
                $(".open_dialog").click(function(e) {
                    e.preventDefault();
                    
                    var linkID = $(this).attr('id')
                    var selector = "target" + linkID;

                    // Define a new observer
                    var obs = new MutationObserver(function(mutations, 
                                                            observer) {
                      // look through all mutations that just occured
                      for(var i=0; i<mutations.length; ++i) {
                        // look through all added nodes of this mutation
                        for(var j=0; j<mutations[i].addedNodes.length; ++j) {
                          // was a child added with the selector ID? 
                          if(mutations[i].addedNodes[j].id == selector) {

                                    displayDialog(linkID);
                          }
                        }
                      }
                    });

                    // have the observer observe the document body 
                    // for changes in children nodes
                    obs.observe($("body").get(0), {
                      childList: true
                    });


                    var $div = $("<div>", {id: selector, class: "dialog", 
                                           style: "display:none;"});

                    var dialogcontent = get_genes(linkID);

                    var output = '<table style="border:none; width:100%;">';
                    for (t in dialogcontent){
                        var descript = get_desc(dialogcontent[t]);
                        output += '''+self.link1+'''}
                    output += '</table>'
                    $div.html(output);
                    $("body").append($div);

                    }); // closes "open_dialog" click event

            }); // closes the function...


            function displayDialog(targetDiv) {

              var name = document.getElementById( targetDiv ).name
              var dialog = $( document.getElementById( 'target' + 
                                                        targetDiv ) ).dialog({
                title: name,
                width: 800
              });
            } 

            function isInArray(value, array) {
              return array.indexOf(value) > -1;
            }

            function get_desc(id){
                var desc = descmap[id]
                return desc
            }

            function get_genes(id){ 
              var slice = annotmap[id];
              var out = [];
              for (g in slice) { 
                if (isInArray(slice[g], genes)){
                  out.push(slice[g]);
                } 
              }
              return out;
            }
            '''+"var genes = "+jgenes+" \n\n "+"</script>"


            js_resources += bigjson



        # pre-process enrichment data for display
        t_bp, t_mf, t_cc, t_kegg = self.process_enrichment_dict(bp, mf, cc,
                                                                kegg, alpha)


        if bp is not None or mf is not None or cc is not None:
            goheader = '''<p style="font-size:20px; font-weight: bold;">
                                                     GO term enrichment</p>'''
        else:
            goheader = ""



        # BIOLOGICAL PROCESS table
        # =====================================================================

        if bp is not None:
            bp_info = '''
            <span style="font-size:16px; font-weight: bold;">
            Biological Process
            </span>

            <a id="displayBP" href="javascript:toggle('BP');">hide</a></br>
            <div id="toggleBP" style="width:1200px; display:block;">

            {% block table1a %}
            {{ tablegobp }}
            {% endblock %}
            <br/>

            </div>
            </br>
            '''

        else:
            bp_info = ""

        # MOLECULAR FUCNTION table
        # =====================================================================

        if mf is not None:
            mf_info = '''
            <span class="secthead">Molecular Function</span>

            <a id="displayMF" href="javascript:toggle('MF');">show</a></br>
            <div id="toggleMF" class="collapsedDiv">

            {% block table1b %}
            {{ tablegomf }}
            {% endblock %}
            <br/>

            </div>
            </br>
            '''
        else:
            mf_info = ""


        # CELLULAR COMPONENT table
        # =====================================================================

        if cc is not None:
            cc_info = '''
            <span class="secthead">Cellular Component</span>

            <a id="displayCC" href="javascript:toggle('CC');">show</a></br>
            <div id="toggleCC" class="collapsedDiv">

            {% block table1c %}
            {{ tablegocc }}
            {% endblock %}
            <br/>
            <br/>

            </div>
            </br>
            '''
        else:
            cc_info = ""


        # KEGG PATHWAYS table
        # =====================================================================

        if kegg is not None:
            kegg_info = '''
            <span class="secthead">KEGG pathways enrichment</span>
            {% block table2 %}
            {{ tablekegg }}
            {% endblock %}
            <br/>
            '''
        else:
            kegg_info = ""

    # -------------------------------------------------------------------------

        template = Template('''<!DOCTYPE html>
        <html lang="en">
            <head>
                <meta charset="utf-8">
                <title>{{ title }}</title>
                {{ js_resources }}
                {{ css_resources }}
                {{ script }}
            </head>
            <body>
                {{ header }}
                
                {{ svgdiv }}
                <br/>

                <br/>
                '''+goheader+bp_info+mf_info+cc_info+kegg_info+
                ''' 
                {% block table3 %}
                {{ annots }}
                {% endblock %}

            </body>
        </html>
        ''')

        html = template.render(js_resources=js_resources,
                               css_resources=css_resources,
                               title = self.title,
                               header = self.process_title(),
                               svgdiv = self.render_svg_inset(),
                               tablegobp = t_bp,
                               tablegomf = t_mf,
                               tablegocc = t_cc,
                               tablekegg = t_kegg,
                               annots = self.render_gene_table(),
                               )
        return html



    def process_title(self):

        """
        NOTE: currently a passthrough placeholder!
        TODO: Refine this method...
        """

        return '<h2>'+self.title+'</h2>'



    def render_svg_inset(self):

        """
        Renders the svg inset tabs on the html report files
        """

        svgdiv = Template('''

            <div id="dagtabs" style="width: 1200px;">

                <ul class="tabs">
                    <li><a href="#view1">Biological Process</a></li>
                    <li><a href="#view2">Molecular Function</a></li>
                    <li><a href="#view3">Cellular Component</a></li>
                </ul>
                <div class="tabcontents">

                    <div id="view1">
                        <object id="bpsvg" type="image/svg+xml" 
                         data="{{ svg['bp'] }}" width="100%" height="600px">
                            Your browser does not support SVG
                        </object>
                    </div>

                    <div id="view2">
                        <object id="mfsvg" type="image/svg+xml" 
                         data="{{ svg['mf'] }}" width="100%" height="600px">
                            Your browser does not support SVG
                        </object>
                    </div>

                    <div id="view3">
                        <object id="ccsvg" type="image/svg+xml" 
                         data="{{ svg['cc'] }}" width="100%" height="600px">
                            Your browser does not support SVG
                        </object>
                    </div>

                </div>
            </div>


            <script>

                var sbp = document.getElementById("bpsvg");
                sbp.addEventListener('load', function(){
                    var panZoomArm = svgPanZoom('#bpsvg', {
                                      zoomEnabled: true,
                                      controlIconsEnabled: true
                                     });
                });

                var smf = document.getElementById("mfsvg");
                smf.addEventListener('load', function(){
                    var panZoomArm = svgPanZoom('#mfsvg', {
                                      zoomEnabled: true,
                                      controlIconsEnabled: true
                                     });
                });

                var scc = document.getElementById("ccsvg");
                scc.addEventListener('load', function(){
                    var panZoomArm = svgPanZoom('#ccsvg', {
                                      zoomEnabled: true,
                                      controlIconsEnabled: true
                                     });
                });

            </script>


        ''')

        div = svgdiv.render(svg=self.svgpaths)
        return div



    # make Jinja2 table template for gene/transcript annotation data/list
    # NOTE: Currently all available annotations per gene/transcript are
    # dumped into a single cell. TODO: ponder the best way to improve that!
    # Maybe a sparse matrix-like table via pandas dataframe? 
    # --------------------------------------------------------------------
    def render_gene_table(self):

        """
        Renders the available annotations for the genes/transcripts being
        plotted into 1st column: gene/transcript identifier; 2nd column:
        all available annotations.
        """

        tablegene = Template('''

        <br/>

            <span class="secthead">
                Genes/transcripts ({{ len1 }}/{{ total }})</span>

        <a id="displayList" href="javascript:toggle('List');">show</a>

        <div id="toggleList" class="collapsedDiv">
            <table class="genedesc">
            <thead>
                <tr>
                    <th scope="col">Identifier</th>
                    <th scope="col">Descriptor</th>
                </tr>
            </thead>
            <tbody>
                {% for id in genedata: %}
                    {% if genedata[id][0].strip() !="": %}
                    <tr>
                        <td>
                        '''+self.link2+'''
                        </td>
                        <td>
                        {{ genedata[id][0].strip() }}
                        </td>
                    </tr>
                    {% endif %}
                {% endfor %}
                <br/>

            </tbody>
        </table>

        </div>

        <h3> No description ({{ len2 }})</h3>
        <p>
        <a id="displayBastards" href="javascript:toggle('Bastards');">show</a>

        <div id="toggleBastards" class="collapsedDiv">
        {% for id in extra %}
            '''+self.link2+'''
            {% if not loop.last %}
                , 
            {% endif %}
        {% endfor %}
        </div>

        ''')


        # CHEEKY HACK
        # ---------------------------------------------------------------------
        naughty = []
        stripped = {}

        for key in self.annots:

            if self.annots[key][0].strip() == "":
                naughty.append(key)
            else:
                stripped.setdefault(key,[]).append(self.annots[key][0].strip())

        lenstrp = len(stripped)
        lennaut = len(naughty)
        # ---------------------------------------------------------------------



        if len(self.annots) > 0:
            table = tablegene.render(genedata=stripped, 
                                     extra=naughty, 
                                     len1=str(lenstrp), 
                                     len2=lennaut, 
                                     total=str(lenstrp+lennaut))
        else:
            table = Template("").render()
            print "uh oh!"
            print "Something went wrong generating the gene list table!"

        return table



    def go_slice(self, enrichRes, alpha):

        """
        Slices up and post-processes individual GO ontology 
        enrichment results for displaying.
        """

        ontslice = enrichRes.loc[enrichRes['elimFisher'] < alpha]
        if len(ontslice) > 0:

            ontslice["Significant"] = ontslice["GO.ID"].map(str) + str("|") + \
                                      ontslice["Term"].map(str) + str("|") + \
                                      ontslice["Significant"].map(str)
            ontslice["Significant"] = ontslice["Significant"].map( 
                lambda x: \
                '''<a id="%s" name="GO: %s" class="open_dialog" 
                href="javascript:void(0)">%s</a> '''% tuple(x.split("|")) )

            ontslice["GO.ID"] = ontslice["GO.ID"].map(
                lambda x: '<a href="http://amigo.geneontology.org/amigo/term/'
                +str(x)+'" target="_blank">'+str(x)+'</a>')

            cols = ['GO.ID', 'Term', 'Annotated', 'Significant', 'elimFisher']
            restable = ontslice[cols].to_html(index=False, 
                                              classes="etables sortable", 
                                              escape=False)
            restable = self.process_go_headers(restable)

            restable = self.process_hidden_keys(restable)

        else:
            restable = self.not_found_response()

        return restable


    # TODO: Evaluate if displaying "No significant enriched terms found!"
    #       is of worth, or just leave blank... **kwargs...it?
    def process_enrichment_dict(self, gobp, gomf, gocc, kegg, alpha):

        """
        Method receives dataframes containing GO/KEGG enrichment results and
        converts them into html tables.
        """

        if gobp is not None:
            gobptable = self.go_slice(gobp, alpha)
        else:
            gobptable = self.not_found_response()


        if gomf is not None:
            gomftable = self.go_slice(gomf, alpha)
        else:
            gomftable = self.not_found_response()


        if gocc is not None:
            gocctable = self.go_slice(gocc, alpha)
        else:
            gocctable = self.not_found_response()


        if kegg is not None:
            
            kegg["Count"] =  kegg["KEGGID"].map(str) + str("|") + \
                             kegg["Term"].map(str) + str("|") + \
                             kegg["Count"].map(str) 
            kegg["Count"] = kegg["Count"].map( 
                lambda x: \
                '''<a id="%s" name="KEGG: %s" class="open_dialog" 
                    href="javascript:void(0)">%s</a> '''% tuple(x.split("|")) )

            kegg["KEGGID"] = kegg["KEGGID"].map(
                lambda x: \
                '<a href="http://www.genome.jp/dbget-bin/www_bget?pathway:map'
                +str(x)+'" target="_blank">'+str(x)+'</a>')

            # changes the column order of the KEGG enrichment results 
            cols = ['KEGGID', 'Term', 'Size', 'Count', 'Pvalue']
            keggtable = kegg[cols].to_html(index=False, 
                                           classes="etables sortable", 
                                           escape=False)

            keggtable = keggtable.replace('<th>KEGGID</th>', 
                                    '<th class="sorttable_nosort">KEGGID</th>')

        else:
            keggtable = self.not_found_response()

        return gobptable, gomftable, gocctable, keggtable


    # render this line when no GO terms found to be enriched 
    def not_found_response(self):

        tablego = Template('<br/>No significant enriched terms found!<br/>')
        table = tablego.render()
        return table

    def process_go_headers(self, table):

        out = table.replace('<th>GO.ID</th>', 
                            '<th class="sorttable_nosort">GO.ID</th>')
        return out


    def process_hidden_keys(self, table):

        """
        Necessary trick to be able to sort exponential
        numbers with the sorttable library.
        """

        getcontext().prec = 100
        patt = re.compile("<td>([0-9-e\.]+)</td>\s+</tr>")

        for m in re.finditer(patt, table):
            d = Decimal(m.group(1))
            s = '{0:f}'.format(d)
            replacement = '<td sorttable_customkey="'+s+'">'+ \
                                    m.group(1)+'</td>\n</tr>\n'
            table = table.replace(m.group(0), replacement)

        return table